﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/04/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Threading.MessagePumping
{
    /// <summary>
    /// Provides a way to synchronously or asynchronously execute a message. 
    /// </summary>
    public interface ISynchronizedMessagePump
    {
        /// <summary>
        /// Gets a value indicating whether the caller can bypass the queue to execute a message.
        /// </summary>
        bool InvokeRequired { get; }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult BeginInvoke(object message);

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult BeginInvoke(object processor, object message);

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult BeginInvoke(object processor, object message, bool enqueue);

        /// <summary>
        /// Waits until the process started by calling <see cref="BeginInvoke"/> completes, and then returns the value generated by the process. 
        /// </summary>
        /// <param name="result">An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling <see cref="BeginInvoke"/>.</param>
        /// <returns>An Object that represents the return value generated by the asynchronous operation.</returns>
        object EndInvoke(IAsyncResult result);

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object Invoke(object message);

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object Invoke(object processor, object message);

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object Invoke(object processor, object message, bool enqueue);
    }
}