﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  06/04/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Threading.MessagePumping
{
    using System.Collections.Concurrent;

    /// <summary>
    /// Manages executing messages on a <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>.
    /// </summary>
    /// <typeparam name="TMessage">The type of messages to execute.</typeparam>
    /// <typeparam name="TResult">The executed message result type.</typeparam>
    public sealed class MessagePump<TMessage, TResult> : ISynchronizedMessagePump, IDisposable
    {
        #region SynchronizedMessagePumpAsyncResult Class
        /// <summary>
        /// Represents the status of a queued message.
        /// </summary>
        private class SynchronizedMessagePumpAsyncResult : IAsyncResult
        {
            #region Disposing
            ~SynchronizedMessagePumpAsyncResult()
            {
                //MS uses this pattern for the ThreadMethodEntry class, should we implement IDisposible?
                if (_WaitHandle != null)
                    _WaitHandle.Close();
            }
            #endregion

            #region Fields
            /// <summary>
            /// Defines the processed message result.
            /// </summary>
            private TResult _Result;
            /// <summary>
            /// Defines a handle to synchronize access.
            /// </summary>
            private readonly ManualResetEvent _WaitHandle;
            /// <summary>
            /// Determines if the message is executed synchronously.
            /// </summary>
            private readonly bool _Synchronous;
            /// <summary>
            /// Determines if the message has been executed.
            /// </summary>
            private bool _Completed;
            #endregion

            #region Properties
            /// <summary>
            /// Gets an exception that occurred while executing the message.
            /// </summary>
            public Exception Exception
            {
                get;
                internal set;
            }

            /// <summary>
            /// Determines if the message has executed.
            /// </summary>
            public bool IsCompleted
            {
                get
                {
                    return _Completed;
                }
                internal set
                {
                    _Completed = value;

                    if (!CompletedSynchronously)
                        _WaitHandle.Set();
                }
            }

            /// <summary>
            /// Gets if the message was executed synchronously.
            /// </summary>
            public bool CompletedSynchronously
            {
                get
                {
                    return IsCompleted && _Synchronous;
                }
            }

            /// <summary>
            /// Gets the executed message result.
            /// </summary>
            public TResult Result
            {
                get
                {
                    if (Exception != null)
                        throw Exception;

                    return _Result;
                }
                internal set
                {
                    _Result = value;
                }
            }

            /// <summary>
            /// Gets the executed message result.
            /// </summary>
            public object AsyncState
            {
                get
                {
                    return Result;
                }
            }

            /// <summary>
            /// Gets a handle to synchronize access.
            /// </summary>
            public WaitHandle AsyncWaitHandle
            {
                get
                {
                    return _WaitHandle;
                }
            }

            /// <summary>
            /// Gets the message to execute.
            /// </summary>
            public TMessage Message
            {
                get;
                private set;
            }

            /// <summary>
            /// Gets the <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> to execute the <see cref="Message"/> with.
            /// </summary>
            public IMessageProcessor<TMessage, TResult> Processor
            {
                get;
                private set;
            }
            #endregion

            /// <summary>
            /// Initializes the <see cref="SynchronizedMessagePumpAsyncResult"/> with a <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and message to execute and a value determining if the message is executed synchronously.
            /// </summary>
            /// <param name="processor">The processor to execute the message on.</param>
            /// <param name="message">The message to execute.</param>
            /// <param name="synchronous">true to synchronously execute the message; otherwise, false.</param>
            public SynchronizedMessagePumpAsyncResult(IMessageProcessor<TMessage, TResult> processor, TMessage message, bool synchronous)
            {
                Processor = processor;
                Message = message;

                _Synchronous = synchronous;
                _WaitHandle = new ManualResetEvent(false);
            }
        }
        #endregion

        #region Fields
        /// <summary>
        /// Defines a synchronization object for the pump state.
        /// </summary>
        private readonly object _SyncObj = new object();
        /// <summary>
        /// Defines an synchronization object for disposing. 
        /// </summary>
        private readonly object _DisposingSyncObj = new object();
        /// <summary>
        /// Defines a collection queued messages.
        /// </summary>
        private readonly ConcurrentQueue<SynchronizedMessagePumpAsyncResult> _Messages = new ConcurrentQueue<SynchronizedMessagePumpAsyncResult>();
        /// <summary>
        /// Defines a handle to suspend the worker thread until messages are available.
        /// </summary>
        private readonly AutoResetEvent _Reset = new AutoResetEvent(false);
        /// <summary>
        /// Defines a worker thread to execute messages on.
        /// </summary>
        private readonly Thread _Thread;
        /// <summary>
        /// Determines if this is currently pumping messages.
        /// </summary>
        private bool _Active;
        /// <summary>
        /// Determines if this has been disposed.
        /// </summary>
        private bool _Disposed;
        /// <summary>
        /// Determines if this is being disposed.
        /// </summary>
        private bool _Disposing;
        #endregion

        #region Properties
        /// <summary>
        /// Gets a value indicating whether the caller can bypass the queue to execute a message.
        /// </summary>
        public bool InvokeRequired
        {
            get
            {
                return Thread.CurrentThread != _Thread;
            }
        }

        /// <summary>
        /// Gets the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}" /> used to process messages.
        /// </summary>
        public IMessageProcessor<TMessage, TResult> DefaultProcessor
        {
            get;
            private set;
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes the <see cref="T:System.Threading.MessagePump`2{TMessage, TResult}"/>.
        /// </summary>
        public MessagePump()
        {
            _Thread = new Thread(RunMessageLoop);
            _Thread.Start();
        }

        /// <summary>
        /// Initializes the <see cref="T:System.Threading.MessagePump`2{TMessage, TResult}"/> with a default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>.
        /// </summary>
        /// <param name="defaultProcessor"></param>
        public MessagePump(IMessageProcessor<TMessage, TResult> defaultProcessor)
            : this()
        {
            DefaultProcessor = defaultProcessor;
        }
        #endregion

        #region Methods
        /// <summary>
        /// Signals the message loop to begin processing.
        /// </summary>
        private void Pulse()
        {
            lock (_SyncObj)
            {
                if (!_Active)
                {
                    _Active = true;
                    _Reset.Set();
                }
            }
        }

        /// <summary>
        /// Executes messages until <see cref="Dispose"/> is called.
        /// </summary>
        private void RunMessageLoop()
        {
            try
            {
                for (; ; ) //loop
                {
                    //if no messages, wait for signal
                    if (_Messages.Count == 0)
                        _Reset.WaitOne();

                    try
                    {
                        do
                        {
                            SynchronizedMessagePumpAsyncResult msg;
                            //attempt to dequeue next message
                            if (_Messages.TryDequeue(out msg))
                            {
                                TMessage message = msg.Message;

                                try
                                {
                                    //execute the message and set the result
                                    msg.Result = (TResult)msg.Processor.Execute(message);
                                }
                                catch (Exception ex) //catch all exceptions (thrown when accessing result)
                                {
                                    //set exception property
                                    msg.Exception = ex;
                                }
                                finally
                                {
                                    //signals WaitHandle
                                    msg.IsCompleted = true;
                                }
                            }
                        }
                        while (_Messages.Count > 0); //loop while messages
                    }
                    finally
                    {
                        //allows Pulse method to signal WaitHandle
                        lock (_SyncObj)
                            _Active = false;
                    }

                    //break if disposing
                    lock (_DisposingSyncObj)
                        if (_Disposing)
                            break;
                }
            }
            finally
            {
                _Reset.Close();
                _Reset.Dispose();
            }
        }
        #endregion

        #region Disposing
        ~MessagePump()
        {
            Dispose(false);
        }

        /// <summary>
        /// Disposes the <see cref="T:System.Threading.MessagePump`2{TMessage, TResult}"/>, waits until any queued messages have been processed.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes unmanaged and optionally, managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            if (disposing)
            {
                bool lockTaken = false;
                Monitor.Enter(_DisposingSyncObj, ref lockTaken);

                try
                {
                    if (_Disposed)
                        return;

                    _Disposing = true;

                    if (InvokeRequired)
                    {
                        Pulse();

                        if (lockTaken)
                        {
                            Monitor.Exit(_DisposingSyncObj);
                            lockTaken = false;
                        }

                        _Thread.Join();
                    }
                }
                finally
                {
                    _Disposed = true;

                    if (lockTaken)
                        Monitor.Exit(_DisposingSyncObj);
                }
            }
            else
            {
                lock (_DisposingSyncObj)
                {
                    _Disposing = true;
                    _Disposed = true;
                }
            }
        }
        #endregion

        #region Interface Implementation
        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult ISynchronizedMessagePump.BeginInvoke(object message)
        {
            return BeginInvoke((TMessage)message);
        }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        public IAsyncResult BeginInvoke(TMessage message)
        {
            return BeginInvoke(DefaultProcessor, message);
        }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult ISynchronizedMessagePump.BeginInvoke(object processor, object message)
        {
            return BeginInvoke((IMessageProcessor<TMessage, TResult>)processor, (TMessage)message);
        }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        public IAsyncResult BeginInvoke(IMessageProcessor<TMessage, TResult> processor, TMessage message)
        {
            return BeginInvoke(processor, message, true);
        }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        IAsyncResult ISynchronizedMessagePump.BeginInvoke(object processor, object message, bool enqueue)
        {
            return BeginInvoke((IMessageProcessor<TMessage, TResult>)processor, (TMessage)message, enqueue);
        }

        /// <summary>
        /// Queues a message to the internal message pump thread to be asynchronously executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/>. 
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling this method.</returns>
        public IAsyncResult BeginInvoke(IMessageProcessor<TMessage, TResult> processor, TMessage message, bool enqueue)
        {
            if (processor == null)
                throw new ArgumentNullException("processor");

            bool invokeRequired = InvokeRequired;
            SynchronizedMessagePumpAsyncResult result = new SynchronizedMessagePumpAsyncResult(processor, message, invokeRequired);

            bool lockTaken = false;
            Monitor.Enter(_DisposingSyncObj, ref lockTaken);

            try
            {
                if (_Disposing || _Disposed)
                    throw new ObjectDisposedException("MessagePump");

                if (invokeRequired || enqueue)
                {
                    _Messages.Enqueue(result);

                    if (invokeRequired)
                        Pulse();
                }
                else
                {
                    if (!enqueue)
                        throw new ArgumentException("Must queue specified message when invoke required.");

                    if (lockTaken)
                    {
                        Monitor.Exit(_DisposingSyncObj);
                        lockTaken = false;
                    }

                    result.Result = processor.Execute(message);
                }
            }
            finally
            {
                if (lockTaken)
                    Monitor.Exit(_DisposingSyncObj);
            }

            return result;
        }

        /// <summary>
        /// Waits until the process started by calling <see cref="BeginInvoke"/> completes, and then returns the value generated by the process. 
        /// </summary>
        /// <param name="result">An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling <see cref="BeginInvoke"/>.</param>
        /// <returns>An Object that represents the return value generated by the asynchronous operation.</returns>
        object ISynchronizedMessagePump.EndInvoke(IAsyncResult result)
        {
            return EndInvoke(result);
        }

        /// <summary>
        /// Waits until the process started by calling <see cref="BeginInvoke"/> completes, and then returns the value generated by the process. 
        /// </summary>
        /// <param name="result">An <see cref="IAsyncResult"/> interface that represents the asynchronous operation started by calling <see cref="BeginInvoke"/>.</param>
        /// <returns>An Object that represents the return value generated by the asynchronous operation.</returns>
        public TResult EndInvoke(IAsyncResult result)
        {
            if (result == null)
                throw new ArgumentNullException("result");

            SynchronizedMessagePumpAsyncResult sync = (SynchronizedMessagePumpAsyncResult)result;

            if (!sync.IsCompleted)
                sync.AsyncWaitHandle.WaitOne();

            return sync.Result;
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object ISynchronizedMessagePump.Invoke(object message)
        {
            return Invoke((TMessage)message);
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the default <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        public TResult Invoke(TMessage message)
        {
            return Invoke(DefaultProcessor, message);
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object ISynchronizedMessagePump.Invoke(object processor, object message)
        {
            return Invoke((IMessageProcessor<TMessage, TResult>)processor, (TMessage)message);
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        public TResult Invoke(IMessageProcessor<TMessage, TResult> processor, TMessage message)
        {
            return Invoke(processor, message, true);
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        object ISynchronizedMessagePump.Invoke(object processor, object message, bool enqueue)
        {
            return Invoke((IMessageProcessor<TMessage, TResult>)processor, (TMessage)message, enqueue);
        }

        /// <summary>
        /// Synchronously queues a message to the internal message pump thread to be executed on the specified <see cref="T:System.Threading.IMessageProcessor`2{TMessage, TResult}"/> and marshals the call to the creating thread.
        /// </summary>
        /// <param name="processor">The processor to execute the message on.</param>
        /// <param name="message">A message to add to the end of the queue.</param>
        /// <param name="enqueue">true to add the message to queue the message; false to immediately execute the message if <see cref="InvokeRequired"/> is false.</param>
        /// <returns>An Object that represents the return value from the message being added, or null if the message has no return value.</returns>
        public TResult Invoke(IMessageProcessor<TMessage, TResult> processor, TMessage message, bool enqueue)
        {
            IAsyncResult result = BeginInvoke(processor, message, enqueue);
            return EndInvoke(result);
        }
        #endregion
    }
}